from unittest import TestCase
from osirisl1services import spatialrotation as sr
import numpy as np


class TestSpatialRotation(TestCase):
    def setUp(self):
        self.q1 = sr.UnitQuaternion(a=0.6119178, b=-0.5256213, c=0.5120781, d=-0.2950506)
        self.q2 = sr.UnitQuaternion(a=0.6118199, b=-0.5257025, c=0.5118889, d=-0.2954371)
        self.q3 = sr.UnitQuaternion(a=-0.6117191, b=0.5257868, c=-0.5117007, d=0.2958215)
        self.q4 = sr.UnitQuaternion(
            a=0.9999986947927778, b=0.00029300037078118816, c=-0.0007884210472700737, d=-0.0013794766310973054)
        self.v1 = sr.Vector(x=1, y=2, z=-3)
        self.v2 = sr.Vector(x=5, y=-6, z=7)

    def test_vector_interpolation(self):
        v = self.v1.interpolate(self.v2, alpha=0.4)
        self.assertAlmostEqual(v.x, 2.6)
        self.assertAlmostEqual(v.y, -1.2)
        self.assertAlmostEqual(v.z, 1.0)

    def test_taitbryanzyx(self):
        tb = sr.TaitBryanZYX(roll=np.radians(0.0337), pitch=np.radians(-0.0903), yaw=np.radians(-0.15810304))
        tbq = tb.to_quaternion()
        self.assertAlmostEqual(tbq.a, self.q4.a)
        self.assertAlmostEqual(tbq.b, self.q4.b)
        self.assertAlmostEqual(tbq.c, self.q4.c)
        self.assertAlmostEqual(tbq.d, self.q4.d)

    def test_mul(self):
        product = self.q1 * self.q2
        expected = sr.UnitQuaternion(
            a=-0.25123293319938, b=-0.6435260302175398, c=0.6263543069428799, d=-0.36116002056564)
        self.assertAlmostEqual(product.a, expected.a)
        self.assertAlmostEqual(product.b, expected.b)
        self.assertAlmostEqual(product.c, expected.c)
        self.assertAlmostEqual(product.d, expected.d)

    def test_mul_with_scalar(self):
        factor = 2.5
        product = self.q1 * factor
        self.assertAlmostEqual(product.a, factor * self.q1.a)
        self.assertAlmostEqual(product.b, factor * self.q1.b)
        self.assertAlmostEqual(product.c, factor * self.q1.c)
        self.assertAlmostEqual(product.d, factor * self.q1.d)

    def test_sub(self):  # Also tests add
        difference = self.q1 - self.q4
        expected = sr.UnitQuaternion(
            a=-0.38808089479277785, b=-0.5259143003707811, c=0.51286652104727, d=-0.29367112336890266)
        self.assertAlmostEqual(difference.a, expected.a)
        self.assertAlmostEqual(difference.b, expected.b)
        self.assertAlmostEqual(difference.c, expected.c)
        self.assertAlmostEqual(difference.d, expected.d)

    def test_eci_to_geo(self):
        eci = sr.ECI(-1122211.53464187, 2871408.28035802, 6250559.77485962)
        expected_ecef = sr.Vector(-1438775.7196152459, 2726585.5331600225, 6250559.7748596193)
        mjd = 52393.37880836018
        ecef = eci.to_geo(mjd)
        np.testing.assert_allclose(expected_ecef, ecef, atol=1e-5)

    def test_slerp_far(self):
        dot = self.q1.dot(self.q4)
        self.assertLess(dot, 0.995)
        self.assertGreater(dot, 0.0)
        far = self.q1.slerp(self.q4, alpha=0.4)
        expected = sr.UnitQuaternion(
            a=0.8539262004096326, b=-0.34576205977074254, c=0.3366261651258157, d=-0.1947856791897259)
        self.assertAlmostEqual(far.a, expected.a)
        self.assertAlmostEqual(far.b, expected.b)
        self.assertAlmostEqual(far.c, expected.c)
        self.assertAlmostEqual(far.d, expected.d)

    def test_slerp_near(self):
        self.assertGreater(self.q1.dot(self.q2), 0.995)
        near = self.q1.slerp(self.q2, alpha=0.4)
        expected = sr.UnitQuaternion(
            a=0.6118786538561652, b=-0.5256537919035789, c=0.5120024315944399, d=-0.2952052066850055)
        self.assertAlmostEqual(near.a, expected.a)
        self.assertAlmostEqual(near.b, expected.b)
        self.assertAlmostEqual(near.c, expected.c)
        self.assertAlmostEqual(near.d, expected.d)

    def test_slerp_switch_handedness_far(self):
        dot = self.q1.dot(-self.q4)
        self.assertGreater(dot, -0.995)
        self.assertLess(dot, 0.0)
        far = self.q1.slerp(-self.q4, alpha=0.4)
        expected = sr.UnitQuaternion(
            a=0.8539262004096326, b=-0.34576205977074254, c=0.3366261651258157, d=-0.1947856791897259)
        self.assertAlmostEqual(far.a, expected.a)
        self.assertAlmostEqual(far.b, expected.b)
        self.assertAlmostEqual(far.c, expected.c)
        self.assertAlmostEqual(far.d, expected.d)

    def test_slerp_switch_handedness_near(self):
        self.assertLess(self.q2.dot(self.q3), -0.995)
        near = self.q2.slerp(self.q3, alpha=0.4)
        expected = sr.UnitQuaternion(
            a=0.6117795905048954, b=-0.5257362290274409, c=0.5118136287883752, d=-0.2955908650756043)
        self.assertAlmostEqual(near.a, expected.a)
        self.assertAlmostEqual(near.b, expected.b)
        self.assertAlmostEqual(near.c, expected.c)
        self.assertAlmostEqual(near.d, expected.d)
